module Statham
  # Internal: Attributes serialized inside JSON.
  class Attribute
    # Internal: Type of the attribute.
    attr_accessor :type
    # Internal: Default value for the attribute.
    attr_accessor :default

    # Internal: Initialize the attribute.
    #
    # options - The hash options used for defining new attributes:
    #           type    - Type of the attribute.
    #           default - Default value for the attribute.
    #
    # Returns new Statham::Attribute object.
    def initialize(options = {})
      @type = options[:type]
      @default = options[:default]
    end

    # Internal: Serializes value to be JSON-compatible.
    #
    # value - Value to serialize.
    #
    # Returns serialized value.
    def serialize(value)
      serialized = value.nil? ? nil : casted_to_type(value)

      serialized.nil? && !@default.nil? ? @default : serialized
    end

    # Internal: Deserializes JSON-compatible value to native one.
    #
    # value - Value to deserialize.
    #
    # Returns deserialized value.
    def deserialize(value, options = {})
      if options[:fallback_to_default]
        value || @default
      else
        value
      end
    end

    protected

    # Internal: Casts value to type defined for the attribute.
    #
    # value - Value to cast type.
    #
    # Returns value with type casted.
    def casted_to_type(value)
      case @type
      when :boolean
        [true, 1, '1', 't', 'T', 'true', 'TRUE', 'on', 'ON'].include?(value)
      when :integer
        value.to_i
      when :array
        value.to_a
      when :hash
        value.to_h
      when :float
        value.to_f
      else
        value.to_s
      end
    end
  end
end
